<?php
/**
* Copyright 2012 Gemeente Schoten (B)
* Licensed under the EUPL, Version 1.1 or - as soon as they will be
* approved by the European Commission - subsequent versions of the
* EUPL (the "Licence");
* You may not use this work except in compliance with the Licence.
*
* A copy of the Licence can be obtained at:
*                http://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in writing,
* software distributed under the Licence is distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the Licence for the specific language governing permissions and
* limitations under the Licence.
*/
/*
* This is (web-)page 2 of a two-page prototype/proof-of-concept
* implementation for a web-based DIV query client.
* It has a rudimentary user interface and only basic security.
* Author: <hjt@ATComputing.nl> September/October 2012, version 1.0
*/

// This page is triggered by a submit of a file with car records
// uploaded from the browser-client via our preceding webpage 'pg1.php'.
$previous_page='/pg1.php';

// Find out which language is preferred by the client's browser:
$LANG=explode(',', $_SERVER['HTTP_ACCEPT_LANGUAGE']);
$LANG=preg_replace('/[;-].*$/', '', $LANG[0]); // isolate first language code
switch($LANG) {
case 'nl': break; // supported language
case 'en': break; // supported language
default:   $LANG='nl'; break;
}
?>
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<?php
echo '<html lang="'.$LANG.'" xml:lang="'.$LANG.'" '
    .'xmlns="http://www.w3.org/1999/xhtml">'."\n";
echo '<head>'."\n";
echo '<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />'."\n";
echo '<title>DIV Client prototype</title>'."\n";

// We cannot yet close the html-<head since more <meta-lines may still follow.
// This implies that we cannot yet open the html-<body either.
// So: buffer these texts to send them out at a later moment:
$headend   = '</head>'."\n";
$bodystart = '<body>'."\n".'<h1>DIV Client</h1>'."\n"
            .'<h2>AT Computing prototype/proof of concept</h2>'."\n";
$inbody = 0;

include_once('WSEConsultTitPlateV02.php');
ini_set('soap.wsdl_cache_enabled', '0');

$today = date('Ymd');           // format becomes YYYYMMDD
$todaycode = date('Ymd-His');   // format becomes YYYYMMDD-HHMMSS

require('./includedir.inc');
require($includedir.'/proxy_opties.inc');
$uniek = rand(1,32767);
$timestamp = time();
require($includedir.'/importiaudit.inc');
$query_client = new WSEConsultTitPlateV02($proxy_opties);
require('./outputdir.inc');

// Go handle the uploaded file (if there is any):
// 'uploadfilename' is (literally so) the name of the text field in the previous page:
$file_handled = 0;
if (isset($_FILES) && isset($_FILES['uploadfilename']) &&
                      is_uploaded_file($_FILES['uploadfilename']['tmp_name'])) {
    // We found an uploaded file; try to open it:
    $inputfilehandle = fopen($_FILES['uploadfilename']['tmp_name'], 'r');
    if ($inputfilehandle == false) {
        switch($LANG) {
        case 'nl': default:
            $msg1 = 'Openen van de inputfile is mislukt. Probeer a.u.b. opnieuw.';
            break;
        case 'en':
            $msg1 = 'Opening of the input file failed. Please retry.';
            break;
        }
        echo '<meta http-equiv="Refresh" content="3; url='.$previous_page.'" />'."\n";
        echo $headend.$bodystart.$msg1."\n"; $inbody++;
        exit('</body></html>'."\n"); // failure exit
    }

    // Before we start reading the uploaded (input) file, we open the primary output
    // file just to be prepared for the output to come:
    // TODO: if multiple clients possibly could upload an input-file within the
    // same second, an output file name-collision check&prevention must be added here.
    $outputfilehandle = fopen($primary_output_dir.'/outfile-'.$todaycode.'.tmp', 'w+');
    if ($outputfilehandle == false) {
        switch($LANG) {
        case 'nl': default:
            $msg1 = 'Cre&euml;ren van de outputfile is mislukt. Probeer a.u.b. opnieuw.';
            break;
        case 'en':
            $msg1 = 'Creating the output file failed. Please retry.';
            break;
        }
        echo '<meta http-equiv="Refresh" content="3; url='.$previous_page.'" />'."\n";
        echo $headend.$bodystart.$msg1."\n"; $inbody++;
        exit('</body></html>'."\n"); // failure exit
    }

    echo $headend.$bodystart; $inbody++;

    // Go process the uploaded file:
    $file_handled =
        process_uploaded_file($inputfilehandle, $outputfilehandle, $today, $todaycode);
    fclose($outputfilehandle);
    fclose($inputfilehandle);

    // Rename the primary output file to its final name, to signal that
    // it is fully ready now, and thus is "fit for consumption":
    rename($primary_output_dir.'/outfile-'.$todaycode.'.tmp',
                    $primary_output_dir.'/outfile-'.$todaycode.'.txt');

    if (!$file_handled) {
        echo '<meta http-equiv="Refresh" content="3; url='.$previous_page.'" />'."\n";
        switch($LANG) {
        case 'nl': default:
            $msg1 = 'De inputfile was leeg. Probeer a.u.b. opnieuw.';
            break;
        case 'en':
            $msg1 = 'The input file was empty. Please retry.';
            break;
        }
        echo $msg1."\n";
        exit('</body></html>'."\n"); // failure exit
    }

    // Distribute copies of the output file to various places:
    distribute_output_files($primary_output_dir, $todaycode, $output_destinations);

} // Handling the uploaded file is ready now.

// Did the user ask for any single numbers (i.e. not from an uploaded file)?
$single_numbers = 0;
foreach($_POST as $postkey => $postvalue) {
    // The single-number fields are named 'single_' followed by a sequence number:
    $single_pattern = '/^single_[0-9]+/';
    if (preg_match($single_pattern, $postkey) && ($postvalue != '')) {
        if (!$inbody) {
            echo $headend.$bodystart; $inbody++;
        }
        $postvalue = trim(strip_tags($postvalue)); // sanitise HTML and trim
        // Would a plausibility check on the car number format be useful here?
        display_start_next_number_message($postvalue, $today);
        $result = do_query($postvalue, $today, 0);
        display_full_result($result);
        $single_numbers++;
    }
} // Handling the single numbers is ready now.

// All input has been handled now. Signal success or failure:
if (!$file_handled && !$single_numbers) {
    switch($LANG) {
    case 'nl': default:
        $msg1 = 'Geen uploaded file en geen individuele nummers gevonden.';
        $msg2 = 'Probeer a.u.b. opnieuw.';
        break;
    case 'en':
        $msg1 = 'No uploaded file nor single numbers found.';
        $msg2 = 'Please retry.';
        break;
    }
    echo '<meta http-equiv="Refresh" content="3; url='.$previous_page.'" />'."\n";
    echo $headend.$bodystart.$msg1.'<br />'.$msg2."\n"; $inbody++;
    exit('</body></html>'."\n"); // failure exit
}
// The program will finish now via the last HTML-statements beyond the functions.
//------------------- end of PHP main part -------------------------------------

// This function reads the uploaded file line by line.
// For each line the database server is contacted, and the answer
// is converted into an output-record, which is written into the
// (primary) output file.
function process_uploaded_file($inputfilehandle, $outputfilehandle, $today, $todaycode)
{
    global $LANG;

    $line_no = 0;
    while ($line = fgets($inputfilehandle, 128)) {
        // We now have one line from the uploaded file: go process it:
        $line_no++;
        $line = trim(strip_tags($line));  // sanitise HTML and trim
        if (strlen($line) == 0) continue; // ignore empty lines
        if ($line[0] == '#')    continue; // lines starting with # are comment
        // We don't (yet) have a check for proper format of an input line.
        $line_array     = explode(';', $line);// split input line in its 3 parts:
        $car_number     = trim($line_array[0]); // car registration no.
        $incident_no    = trim($line_array[1]); // incident (administrative) no.
        $incident_date  = trim($line_array[2]); // date of incident
        // Would a plausibility check on the car number format be useful here?
        // Same question for $incident_date format....
        display_start_next_number_message($car_number, $incident_date);
        // Contact the DIV server:
        $result    = do_query($car_number, $incident_date, 1);
        display_limited_result($result);
        // Convert DIV server's answer into an output record:
        $outrecord = compose_output_record($result, $incident_no,
                                             $incident_date, $today);
        // Write the output record to the output file, with CR+LF appended:
        fputs($outputfilehandle, $outrecord."\r\n");
        fflush($outputfilehandle); // this is a safety precaution
    }
    return($line_no != 0 ? 1 : 0); // 1 for success, 0 for failure
}

// This function queries the database server:
// $car_number: car registration no.
// $incident_date: who owned the car at that date.
// flag==1 for query from uploaded file line
// flag==0 for incidental query (single number field on previous webpage)
// Return value of this function is the request-answer-datastructure
function do_query($car_number, $incident_date, $flag)
{
    global $query_client, $importiaudit, $LANG;

    // This determines the language-code of the request to the DIV.
    // Three numerical values only are supported (no English!).
    // No translation of texts needed at this point in the program.
    switch($LANG) {
    case 'fr': $languagecode = 1; break;
    case 'de': $languagecode = 3; break;
    case 'nl': default: $languagecode = 2; break;
    }

    $importikeys = new ImportIKeys(
        $car_number,         // PlateNr
        '',                  // Vin
        '',                  // Unifier
        $languagecode,       // LanguageCode
        $incident_date       // IncidentDate
    );

    // $timestamp and $uniq members in $importiaudit are for auditing by
    // the DIV only. Within one single batch of requests they may stay
    // at the same value throughout.
    // It may be wise to log these values on the local computer as well...
    // Note: we assume that the DIV has its own precautions against SQL-insertion
    // attacks and the like. It's their database....
    $request = new request($importiaudit, $importikeys);
    $fmv     = new FmvConsultTitPlateV02($request);
    // Contact the DIV database server now:
    $result = $query_client->FmvConsultTitPlateV02($fmv);
    // Debug print to show the entire result-record as came from the database:
    // echo '<p>TEST: result is: <br />'; print_r($result); echo '</p>'; flush();
    // We might add a check on the   $result->error   field here.

    return($result);
}

// This function converts a '$result' answer from the database server
// into an output record to be stored in the output file.
// That output record is the return value of this function.
function compose_output_record($result, $incident_no, $incident_date, $today)
{
    global $LANG;

    // Start composing an output line:
    $outrecord = '';

    // -----------------------------------------------------------
    // The processing steps for each field are identical:
    // (1) trailing and leading whitespace is trimmed off
    // (2) the remainder is cut to the allowed width as a maximum
    // (3) space-padding is done to fill the allowed width exactly
    // (4) the field is appended, with ';' separator, to the output-line
    //
    // Appending a ';' is i.m.h.o. not a wise decision. This assumes that
    // the text content inside the record when sent from the DIV will never
    // contain a ';' character. I doubt whether this can be guaranteed.
    // -----------------------------------------------------------
    // The fields in the output line are fixed width *and* separated by ';'
    // Note: the server delivers UTF-8 encoding for text fields.
    // The fixed width of the fields is counted in bytes, which is not
    // necessarily equal to the number of printable characters (UTF-8 uses
    // 2 bytes for most accented characters related to the Latin alphabet).
    // -----------------------------------------------------------

    // Number plate / Car's registration number
    $width = 10;
    $field = $result->response->ExportIKeys->PlateNr;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // No. of "vaststelling/constatation"
    $width = 6;
    $field = $incident_no;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Date of "vaststelling/constatation" YYYYMMDD
    $width = 8;
    $field = $incident_date;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // StatusCode
    $width = 3;
    $field = $result->response->ExportIRegistration->StatusCode;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // StatusLabel
    $width = 30;
    $field = $result->response->ExportIRegistration->StatusLabel;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Last Name of the Titular
    $width = 160;
    $field = $result->response->ExportITitular->LastName;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // First prenom of the Titular
    $width = 50;
    $field = $result->response->ExportITitular->FirstName1;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Second prenom of the Titular
    $width = 50;
    $field = $result->response->ExportITitular->FirstName2;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Third prenom of the Titular
    $width = 50;
    $field = $result->response->ExportITitular->FirstName3;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Street name
    $width = 100;
    $field = $result->response->ExportITitular->StreetName;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // House no.
    $width = 11;
    $field = $result->response->ExportITitular->HouseNr;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Box no.
    $width = 11;
    $field = $result->response->ExportITitular->BoxNr;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Postal Code
    $width = 4;
    $field = $result->response->ExportITitular->PostalCode;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Nis_no.
    $width = 5;
    $field = $result->response->ExportITitular->NisNr;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // National ID
    $width = 11;
    $field = $result->response->ExportITitular->NationalId;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Start situation date (i.e. date of last modification)
    // Server delivers this in format YYYYMMDD and we leave it that way
    $width = 8;
    $field = $result->response->ExportIRegistration->StartSituationDate;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // City name
    $width = 100;
    $field = $result->response->ExportITitular->CityName;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // VIN (Vehicle Identification Number)
    $width = 17;
    $field = $result->response->ExportIKeys->Vin;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Company no.
    $width = 10;
    $field = $result->response->ExportITitular->CompanyNr;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // Date of today YYYYMMDD
    $width = 8;
    $field = $today;
    $field = substr(trim($field), 0, $width);
    $field = $field.str_repeat(' ', $width - strlen($field));
    $outrecord .= $field.';';

    // -----------------------------------------------------------
    //     No CR and/or LF added to the outputrecord (yet).
    // -----------------------------------------------------------

    // a few debug-print helper routines:
    // echo '<p>Record length became: '.strlen($outrecord).'</p>';
    // $displayrecord = htmlspecialchars($outrecord, ENT_COMPAT, 'UTF-8');
    // $displayrecord = str_replace(' ', '&nbsp;', $displayrecord);
    // echo '<p>'.$displayrecord.'</p>';

    return($outrecord);
}

// This function displays a message on the user's browser screen
// to show that we started the query for yet another number
function display_start_next_number_message($car_number, $incident_date)
{
    global $LANG;

    switch($LANG) {
    case 'nl': default:
        $msg1 = 'We gaan nu vragen om autonummer';
        $msg2 = 'op datum';
        break;
    case 'en':
        $msg1 = 'Start of request for registration no.';
        $msg2 = 'on date';
        break;
    }
    echo '<p>&nbsp;<br />';
    echo $msg1.' '.$car_number.' '.$msg2.' '.$incident_date.':';
    echo '</p>'."\n";
    flush(); // Simple output flushing towards user's browser
}

// This function displays a limited version of one query result
// (i.e. for one car registration no.) on the user's browser screen.
function display_limited_result($result)
{
    global $LANG;

    switch($LANG) {
    case 'nl': default:
        $msg1 = 'Die auto was op die datum van';
        break;
    case 'en':
        $msg1 = 'On that date the car was owned by';
        break;
    }
    $ownerlastname   = htmlspecialchars($result->response->ExportITitular->LastName);
    $ownerfirstname1 = htmlspecialchars($result->response->ExportITitular->FirstName1);
    $errorcode       = htmlspecialchars($result->response->ExportIMessage->Code);
    $errormessage    = htmlspecialchars($result->response->ExportIMessage->Label);

    if (!$errorcode) {
        $displayerrormessage = '';
    } else {
        $displayerrormessage = "\n".'<br />DIV errorcode '.$errorcode.'= '.$errormessage;
    }
    echo '<p>'.$msg1.': '.$ownerlastname.', '.$ownerfirstname1
        .$displayerrormessage.'</p>'."\n";
    flush(); // Simple output flushing towards user's browser
}

// This function displays a full blown version of one query result
// (i.e. for one car registration no.) on the user's browser screen.
// For full details of the DIV's response message format please consult the
// ./wsdl_pro.xml file in this directory. Go to the line containing
// <xsd:element name="response"> and from there on observe the hierarchical
// sequence of the <xsd:element name=...> lines.
// The next level of hierarchy within this response-element has following elements:
// ExportIAudit, ExportIKeys, ExportIMessage, ExportITitular, ExportIRegistration
function display_full_result($result)
{
    global $LANG;
    
    $platenr = htmlspecialchars($result->response->ExportIKeys->PlateNr);
    $vin     = htmlspecialchars($result->response->ExportIKeys->Vin);
    $statuscode  = htmlspecialchars($result->response->ExportIRegistration->StatusCode);
    $statuslabel = htmlspecialchars($result->response->ExportIRegistration->StatusLabel);
    $startdate   = htmlspecialchars($result->response->ExportIRegistration->StartSituationDate);
    $ownerlastname   = htmlspecialchars($result->response->ExportITitular->LastName);
    $ownerfirstname1 = htmlspecialchars($result->response->ExportITitular->FirstName1);
    $ownerfirstname2 = htmlspecialchars($result->response->ExportITitular->FirstName2);
    $ownerfirstname3 = htmlspecialchars($result->response->ExportITitular->FirstName3);
    $ownercompanyno  = htmlspecialchars($result->response->ExportITitular->CompanyNr);
    $ownernationalid = htmlspecialchars($result->response->ExportITitular->NationalId);
    $streetname      = htmlspecialchars($result->response->ExportITitular->StreetName);
    $houseno         = htmlspecialchars($result->response->ExportITitular->HouseNr);
    $boxno           = htmlspecialchars($result->response->ExportITitular->BoxNr);
    $postalcode      = htmlspecialchars($result->response->ExportITitular->PostalCode);
    $cityname        = htmlspecialchars($result->response->ExportITitular->CityName);
    $nisno           = htmlspecialchars($result->response->ExportITitular->NisNr);
    $errorcode       = htmlspecialchars($result->response->ExportIMessage->Code);
    $errormessage    = htmlspecialchars($result->response->ExportIMessage->Label);

    // On many of the above items the htmlspecialchars() treatment is overkill
    // because those fields never would contain non-ASCII characters. But we have
    // left this in just to be sure, until a formal spec from the DIV clarifies
    // which character encoding would be possible in each field.

    echo '<p>';
    // For the time being we don't translate this but mostly use the DIV element names:

    // Some fields in the DIV-answer may or may not have content. If there is
    // content it must be displayed, wrapped in some explanatory label text.
    // For these fields we build a $display...message variable which either holds
    // this combination of field content+label text, or is an empty string.
    // This $display...message then serves as a building block in the final output.

    if ($vin) $displayvinmessage = '&nbsp;&nbsp;(VIN '.$vin.')&nbsp;';
    else      $displayvinmessage = '';
    echo $platenr.$displayvinmessage
        .'&nbsp;&nbsp;&nbsp;Status:&nbsp;'.$statuscode.'&nbsp;&nbsp;'.$statuslabel
        .';&nbsp;&nbsp;&nbsp;StartSituationDate:&nbsp;'.$startdate
        .'<br />'."\n";

    echo $ownerlastname.', '.$ownerfirstname1.' '.$ownerfirstname2.' '.$ownerfirstname3
        .'<br />'."\n";

    if ($ownernationalid) $displaynationalidmessage     = 'NationalId: '.$ownernationalid;
    else                  $displaynationalidmessage     = '';
    if ($ownercompanyno)  $displayownercompanynomessage = 'CompanyID: '.$ownercompanynr;
    else                  $displayownercompanynomessage = '';
    if ($ownernationalid && $ownercompanyno) $displayseparator = ';&nbsp;';
    else                                     $displayseparator = '';
    if ($ownernationalid || $ownercompanyno) {
       echo $displaynationalidmessage.$displayseparator
           .$displayownercompanynomessage.'<br />'."\n";
    }

    if ($boxno) $displayboxnomessage = '&nbsp;(Box '.$boxno.')';
    else        $displayboxnomessage = '';
    echo $streetname.' '.$houseno.$displayboxnomessage.'<br />'."\n";

    if ($nisno) $displaynisnomessage = '&nbsp;(NisNr: '.$nisno.')';
    else        $displaynisnomessage = '';
    echo $postalcode.' '.$cityname.$displaynisnomessage."\n";

    if ($errorcode) $displayerrormessage = "\n".'<br />DIV errorcode '
                                          .$errorcode.'= '.$errormessage;
    else            $displayerrormessage = '';
    echo $displayerrormessage.'</p>';
    flush(); // Simple output flushing towards user's browser
}

// This function puts copies of the primary output file to the proper places
// in the file system, as parameterised by the 'outputdir.inc' include-file:
function distribute_output_files($primary_output_dir, $todaycode, $output_destinations)
{
    global $LANG;

    if (isset($output_destinations)) {
        for ($i=0; $i<count($output_destinations); $i++) {
            $outdest = $output_destinations[$i];
            switch($outdest[0]) { // switch on code field
            case 1: // code 1: use PHP copy() followed by rename():
                // We use 'copy' followed by 'rename' to avoid the possibility that the
                // consumer picks up an incomplete file which is still in the process
                // of being filled. As long as the file name ends on .tmp the consumer
                // should ignore it! File names ending in .txt are "fit for consumption".
                // Note that this relies upon the 'rename' operation being atomic.
                // This requires both old and new name to be within the same disk partition.
                // Just to be sure we even stay within the same directory.
                copy($primary_output_dir.'/outfile-'.$todaycode.'.txt',
                                                      $outdest[1].$todaycode.'.tmp');
                rename($outdest[1].$todaycode.'.tmp', $outdest[1].$todaycode.'.txt');
                break;
            default: echo '<br />output_destinations code number'
                         .$dest[0].' not implemented'."\n";
                break;
            }
        }
    }   
}

?>
<!-- the end -->
</body>
</html>
